import * as dotenv from 'dotenv';
import * as bcrypt from 'bcrypt';
import {Guid} from 'guid-typescript';

dotenv.config();

const mariadb = require('mariadb');
const prod_pool = mariadb.createPool({
    host: process.env.DB_HOST,
    user: process.env.DB_USER,
    password: process.env.DB_PASSWORD,
    database: process.env.PARTYPLANER_PROD_DATABASE,
    connectionLimit: 5
});
const dev_pool = mariadb.createPool({
    host: process.env.DB_HOST,
    user: process.env.DB_USER,
    password: process.env.DB_PASSWORD,
    database: process.env.PARTYPLANER_DEV_DATABASE,
    connectionLimit: 5
});

/**
 * Fetches all usernames and emails from the database and returns them in an object
 * @param useDev If the dev or prod database should be used
 * @return any An object with a list of usernames and emails
 */
export const getExistingUsernamesAndEmails = async (useDev: boolean): Promise<any> => {
    let conn;
    try {
        if (useDev) {
            conn = await dev_pool.getConnection();
        } else {
            conn = await prod_pool.getConnection();
        }

        const rows = await conn.query('SELECT username, email FROM users');

        let usernames: string[] = [];
        let emails: string[] = [];

        for (let row in rows) {
            if (row !== 'meta') {
                usernames.push(rows[row].username);
                emails.push(rows[row].email);
            }
        }

        return {
            'usernames': usernames,
            'emails': emails
        };
    } catch (err) {
        throw err;
    } finally {
        if (conn) {
            conn.end();
        }
    }
};

/**
 * Used in the registerUser and loginUser methods as return value
 */
export interface Session {
    userId: string;
    sessionId: string;
    sessionKey: string;
}

/**
 * Creates a new user and a new session
 * @param useDev If the dev or prod database should be used
 * @param username The username of the new user
 * @param email The email address of the new user
 * @param firstName The first name of the new user
 * @param lastName The last name of the new user
 * @param password The password of the new user
 * @param ip The IP Address of the new user
 * @param deviceInfo The user agent of the new user
 */
export const registerUser = async (useDev: boolean, username: string, email: string, firstName: string, lastName: string, password: string, ip: string, deviceInfo: string): Promise<Session> => {
    let conn;
    try {
        if (useDev) {
            conn = await dev_pool.getConnection();
        } else {
            conn = await prod_pool.getConnection();
        }

        const pwHash = bcrypt.hashSync(password, 10);
        const sessionKey = Guid.create().toString();
        const sessionKeyHash = bcrypt.hashSync(sessionKey, 10);
        const verifyEmailCode = Guid.create().toString();

        // Create user
        const userQuery = 'INSERT INTO users (username, email, first_name, last_name, password_hash, verify_email_code, registration_date, last_login) VALUES (?, ?, ?, ?, ?, ?, NOW(), NOW()) RETURNING user_id';
        const userIdRes = await conn.query(userQuery, [username, email, firstName, lastName, pwHash, verifyEmailCode]);
        await conn.commit();

        // Get user id of the created user
        let userId: number = -1;
        for (const row in userIdRes) {
            if (row !== 'meta' && userIdRes[row].user_id != null) {
                userId = userIdRes[row].user_id;
            }
        }

        let deviceType = 'unknown';
        if (deviceInfo.includes('PartyPlaner') || deviceInfo.includes('Dart')) {
            deviceType = 'mobile';
        } else {
            deviceType = 'desktop';
        }

        // Create session
        const sessionQuery = 'INSERT INTO sessions (user_id, session_key_hash, created_Date, last_login, valid_until, valid_days, last_ip, device_info, type) VALUES (?,?,NOW(),NOW(),DATE_ADD(NOW(), INTERVAL 365 DAY),365,?, ?, ?) RETURNING session_id';
        const sessionIdRes = await conn.query(sessionQuery, [userId, sessionKeyHash, ip, deviceInfo, deviceType]);
        await conn.commit();

        // Get session id of the created session
        let sessionId: number = -1;
        for (const row in sessionIdRes) {
            if (row !== 'meta' && sessionIdRes[row].session_id != null) {
                sessionId = sessionIdRes[row].session_id;
            }
        }

        return {
            'userId': userId.toString(),
            'sessionId': sessionId.toString(),
            'sessionKey': sessionKey
        };
    } catch (err) {
        throw err;
    } finally {
        if (conn) {
            conn.end();
        }
    }
};

/**
 * Checks credentials of a user and creates a new session if they are correct
 * @param useDev If the dev or prod database should be used
 * @param username The username of the new user
 * @param email The email address of the new user
 * @param password The password of the new user
 * @param ip The IP Address of the new user
 * @param deviceInfo The user agent of the new user
 */
export const loginUser = async (useDev: boolean, username: string, email: string, password: string, ip: string, deviceInfo: string): Promise<Session> => {
    let conn;
    try {
        if (useDev) {
            conn = await dev_pool.getConnection();
        } else {
            conn = await prod_pool.getConnection();
        }

        let query_result;

        // Get the saved hash
        if (username !== '') {
            query_result = await conn.query('SELECT user_id, password_hash FROM users WHERE username = ?', username);
        } else {
            query_result = await conn.query('SELECT user_id, password_hash FROM users WHERE email = ?', email);
        }

        let passwordHash: string = '';
        let userId: string = '';

        for (let row in query_result) {
            if (row !== 'meta') {
                passwordHash = query_result[row].password_hash;
                userId = query_result[row].user_id;
            }
        }

        // Wrong password
        if (!bcrypt.compareSync(password, passwordHash)) {
            return {} as Session;
        }

        // Update user last login
        await conn.query('UPDATE users SET last_login = NOW() WHERE user_id = ?', userId);
        await conn.commit();

        // Create session
        const sessionKey = Guid.create().toString();
        const sessionKeyHash = bcrypt.hashSync(sessionKey, 10);

        let deviceType = 'unknown';
        if (deviceInfo.includes('PartyPlaner') || deviceInfo.includes('Dart')) {
            deviceType = 'mobile';
        } else {
            deviceType = 'desktop';
        }

        const sessionQuery = 'INSERT INTO sessions (user_id, session_key_hash, created_Date, last_login, valid_until, valid_days, last_ip, device_info, type) VALUES (?,?,NOW(),NOW(),DATE_ADD(NOW(), INTERVAL 365 DAY),365,?, ?, ?) RETURNING session_id';
        const sessionIdRes = await conn.query(sessionQuery, [userId, sessionKeyHash, ip, deviceInfo, deviceType]);
        await conn.commit();

        // Get session id of the created session
        let sessionId: number = -1;
        for (const row in sessionIdRes) {
            if (row !== 'meta' && sessionIdRes[row].session_id != null) {
                sessionId = sessionIdRes[row].session_id;
            }
        }

        return {
            'userId': userId.toString(),
            'sessionId': sessionId.toString(),
            'sessionKey': sessionKey
        };

    } catch (err) {
        throw err;
    } finally {
        if (conn) {
            conn.end();
        }
    }
};

/**
 * Used in the checkUsernameAndEmail method as return value
 */
export interface Status {
    hasProblems: boolean;
    messages: string[];
    status: string[];
}

/**
 * Checks if the given username and email are valid and not yet used by another user
 * @param useDev If the dev database has to be used
 * @param username The username to check
 * @param email The email to check
 */
export const checkUsernameAndEmail = async (useDev: boolean, username: string, email: string): Promise<Status> => {
    let conn;
    try {
        if (useDev) {
            conn = await dev_pool.getConnection();
        } else {
            conn = await prod_pool.getConnection();
        }
        const usernameQuery = 'SELECT username FROM users WHERE username = ?';
        const emailQuery = 'SELECT email FROM users WHERE email = ?';
        const usernameRes = await conn.query(usernameQuery, username);
        const emailRes = await conn.query(emailQuery, email);

        let res: Status = {
            hasProblems: false,
            messages: [],
            status: []
        };

        const usernameRegex = RegExp('^[a-zA-Z0-9\\-\\_]{4,20}$'); // Can contain a-z, A-Z, 0-9, -, _ and has to be 4-20 chars long
        if (!usernameRegex.test(username)) {
            // Username doesn't match requirements
            res.hasProblems = true;
            res.messages.push('The username contains illegal characters or is not the allowed size');
            res.status.push('INVALID_USERNAME');
        }

        const emailRegex = RegExp('^[a-zA-Z0-9\\-\\_.]{1,30}\\@[a-zA-Z0-9\\-.]{1,20}\\.[a-z]{1,20}$'); // Normal email regex, user@plutodev.de
        if (!emailRegex.test(email)) {
            // Username doesn't match requirements
            res.hasProblems = true;
            res.messages.push('The email contains illegal characters or is not the allowed size');
            res.status.push('INVALID_EMAIL');
        }

        if (usernameRes.length > 0) {
            // Username is a duplicate
            res.hasProblems = true;
            res.messages.push('This username is already in use');
            res.status.push('DUPLICATE_USERNAME');
        }

        if (emailRes.length > 0) {
            // Email is a duplicate
            res.hasProblems = true;
            res.messages.push('This email is already in use');
            res.status.push('DUPLICATE_EMAIL');
        }

        return res;

    } catch (err) {
        throw err;
    } finally {
        if (conn) {
            conn.end();
        }
    }
};
